'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var reduxSaga = require('redux-saga');
var ramda = require('ramda');
var React = require('react');
var React__default = _interopDefault(React);
var kReducer = require('k-reducer');
var effects = require('redux-saga/effects');

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) {
  try {
    var info = gen[key](arg);
    var value = info.value;
  } catch (error) {
    reject(error);
    return;
  }

  if (info.done) {
    resolve(value);
  } else {
    Promise.resolve(value).then(_next, _throw);
  }
}

function _asyncToGenerator(fn) {
  return function () {
    var self = this,
        args = arguments;
    return new Promise(function (resolve, reject) {
      var gen = fn.apply(self, args);

      function _next(value) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value);
      }

      function _throw(err) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err);
      }

      _next(undefined);
    });
  };
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _objectSpread(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    var ownKeys = Object.keys(source);

    if (typeof Object.getOwnPropertySymbols === 'function') {
      ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) {
        return Object.getOwnPropertyDescriptor(source, sym).enumerable;
      }));
    }

    ownKeys.forEach(function (key) {
      _defineProperty(target, key, source[key]);
    });
  }

  return target;
}

function _inheritsLoose(subClass, superClass) {
  subClass.prototype = Object.create(superClass.prototype);
  subClass.prototype.constructor = subClass;
  subClass.__proto__ = superClass;
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread();
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  }
}

function _iterableToArray(iter) {
  if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter);
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance");
}

var matchOperatorsRe = /[|\\{}()[\]^$+*?.]/g;

var escapeStringRegexp = function escapeStringRegexp(str) {
  if (typeof str !== 'string') {
    throw new TypeError('Expected a string');
  }

  return str.replace(matchOperatorsRe, '\\$&');
};

var createSetContextWarning = function createSetContextWarning(ctx, props) {
  return (ctx ? ctx + '.' : '') + 'setContext(props): argument ' + props + ' is not a plain object';
};

function remove(array, item) {
  var index = array.indexOf(item);

  if (index >= 0) {
    array.splice(index, 1);
  }
}

var matcher = function matcher(pattern) {
  var regexp = new RegExp("^" + escapeStringRegexp(pattern) + "\\.(.+)");
  return function (action) {
    if (action.type === pattern || pattern === '') {
      return {
        id: pattern,
        wrap: function wrap(type) {
          return type;
        },
        unwrappedType: action.type,
        args: {}
      };
    } else {
      var match = action.type.match(regexp);

      if (match) {
        var unwrappedType = match[1];
        return {
          id: pattern + ".",
          wrap: function wrap(type) {
            return pattern + "." + type;
          },
          unwrappedType: unwrappedType,
          args: {}
        };
      } else {
        return false;
      }
    }
  };
};

function emitter() {
  var subscribers = [];

  var subscribe = function subscribe(prefix) {
    return function (sub) {
      subscribers.push({
        prefix: prefix,
        sub: sub,
        matcher: matcher(prefix)
      });
      return function () {
        return remove(subscribers, sub);
      };
    };
  };

  function emit(item) {
    var arr = subscribers.slice();

    for (var i = 0, len = arr.length; i < len; i++) {
      var entry = arr[i];
      var match = entry.matcher(item);

      if (match) {
        arr[i].sub(_objectSpread({}, item, {
          type: match.unwrappedType
        }));
      }
    }
  }

  return {
    subscribe: subscribe,
    emit: emit
  };
}

function check(value, predicate, error) {
  if (!predicate(value)) {
    //log('error', 'uncaught at check', error);
    throw new Error(error);
  }
}

function sagaMiddlewareFactory() {
  var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var _ref$context = _ref.context,
      context = _ref$context === undefined ? {} : _ref$context,
      options = ramda.omit(['context'], _ref);
  var sagaMonitor = options.sagaMonitor,
      logger = options.logger,
      onError = options.onError;

  if (ramda.is(Function, options)) {
    throw new Error("You passed a function to the Saga middleware. You are likely trying to start a Saga by directly passing it to the middleware. This is no longer possible starting from 0.10.0. To run a Saga, you must do it dynamically AFTER mounting the middleware into the store.\n        Example:\n          import createSagaMiddleware from 'redux-saga'\n          ... other imports\n\n          const sagaMiddleware = createSagaMiddleware()\n          const store = createStore(reducer, applyMiddleware(sagaMiddleware))\n          sagaMiddleware.run(saga, ...args)\n      ");
  }

  if (logger && !ramda.is(Function, logger)) {
    throw new Error('`options.logger` passed to the Saga middleware is not a function!');
  }

  if (options.onerror) {
    throw new Error('`options.onerror` was removed. Use `options.onError` instead.');
  }

  if (onError && !ramda.is(Function, onError)) {
    throw new Error('`options.onError` passed to the Saga middleware is not a function!');
  }

  if (options.emitter && !ramda.is(Function, options.emitter)) {
    throw new Error('`options.emitter` passed to the Saga middleware is not a function!');
  }

  function sagaMiddleware(_ref2) {
    var _getState = _ref2.getState,
        dispatch = _ref2.dispatch;
    var sagaEmitter = emitter();
    sagaEmitter.emit = (options.emitter || ramda.identity)(sagaEmitter.emit);

    sagaMiddleware.run = function (prefix, saga) {
      for (var _len = arguments.length, args = new Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
        args[_key - 2] = arguments[_key];
      }

      return reduxSaga.runSaga.apply(void 0, [{
        context: context,
        subscribe: sagaEmitter.subscribe(prefix),
        dispatch: prefix ? function (action) {
          return dispatch(_objectSpread({}, action, {
            type: prefix + "." + action.type
          }));
        } : dispatch,
        getState: function getState(x) {
          return console.log('qqqqqqqq', x) || ramda.compose(prefix ? ramda.path(prefix.split('.')) : ramda.identity, _getState)(x);
        },
        sagaMonitor: sagaMonitor,
        logger: logger,
        onError: onError
      }, saga].concat(args));
    };

    return function (next) {
      return function (action) {
        if (sagaMonitor && sagaMonitor.actionDispatched) {
          sagaMonitor.actionDispatched(action);
        }

        var result = next(action); // hit reducers

        sagaEmitter.emit(action);
        return result;
      };
    };
  }

  sagaMiddleware.run = function () {
    throw new Error('Before running a Saga, you must mount the Saga middleware on the Store using applyMiddleware');
  };

  sagaMiddleware.setContext = function (props) {
    check(props, ramda.is(Object), createSetContextWarning('sagaMiddleware', props));
    Object.assign(context, props);
  };

  return sagaMiddleware;
}

var sagaMiddleware = sagaMiddlewareFactory();

/**
 * https://github.com/reduxjs/react-redux
 */
function bindActionCreator(actionCreator, dispatch) {
  return function () {
    return dispatch(actionCreator.apply(this, arguments));
  };
}
/**
 * Turns an object whose values are action creators, into an object with the
 * same keys, but with every function wrapped into a `dispatch` call so they
 * may be invoked directly. This is just a convenience method, as you can call
 * `store.dispatch(MyActionCreators.doSomething())` yourself just fine.
 *
 * For convenience, you can also pass an action creator as the first argument,
 * and get a dispatch wrapped function in return.
 *
 * @param {Function|Object} actionCreators An object whose values are action
 * creator functions. One handy way to obtain it is to use ES6 `import * as`
 * syntax. You may also pass a single function.
 *
 * @param {Function} dispatch The `dispatch` function available on your Redux
 * store.
 *
 * @returns {Function|Object} The object mimicking the original object, but with
 * every action creator wrapped into the `dispatch` call. If you passed a
 * function as `actionCreators`, the return value will also be a single
 * function.
 */


function bindActionCreators(actionCreators, dispatch) {
  if (typeof actionCreators === 'function') {
    return bindActionCreator(actionCreators, dispatch);
  }

  if (typeof actionCreators !== 'object' || actionCreators === null) {
    throw new Error("bindActionCreators expected an object or a function, instead received " + (actionCreators === null ? 'null' : typeof actionCreators) + ". " + "Did you write \"import ActionCreators from\" instead of \"import * as ActionCreators from\"?");
  }

  var keys = Object.keys(actionCreators);
  var boundActionCreators = {};

  for (var i = 0; i < keys.length; i++) {
    var key = keys[i];
    var actionCreator = actionCreators[key];

    if (typeof actionCreator === 'function') {
      boundActionCreators[key] = bindActionCreator(actionCreator, dispatch);
    }
  }

  return boundActionCreators;
}

var defaultContextValue = {
  scope: [],
  assocReducer: function assocReducer() {},
  dissocReducer: function dissocReducer() {},
  dispatch: function dispatch() {},
  runSaga: function runSaga() {},
  getState: function getState() {},
  subscribe: function subscribe() {},
  supplied: false
};
var KLogicContext = React.createContext(defaultContextValue);

function KLogicProvider(_ref) {
  var store = _ref.store,
      runSaga = _ref.runSaga,
      staticReducer = _ref.staticReducer,
      children = _ref.children;

  var _useState = React.useState(defaultContextValue),
      context = _useState[0],
      setContext = _useState[1];

  var reducersTree = React.useRef({});
  var assocReducer = React.useCallback(function (rPath, reducer) {
    if (ramda.path(rPath, reducersTree.current)) {
      console.error('additional scope is required for: ', rPath);
    } else {
      var newTree = ramda.assocPath(rPath, reducer, reducersTree.current);
      reducersTree.current = newTree;
      var treeReducer = kReducer.fromTree(newTree);
      var finalReducer = staticReducer ? kReducer.composeReducers(staticReducer, treeReducer) : treeReducer;
      store.replaceReducer(finalReducer); //setState(store.getState());
    }
  }, [store]);
  var dissocReducer = React.useCallback(function (rPath) {
    console.log('dissocReducer');
    /*
    if (!path(rPath, reducersTree.current)) {
      console.warning('additional scope is required for: ', rPath);
    } else {
      const newTree = dissocPath(rPath, reducersTree.current);
      reducersTree.current = newTree;
      store.replaceReducer(fromTree(newTree));
      setContext({...context, state: store.getState()});
    }
    */
  }, [store]);
  var contextValue = React.useMemo(function () {
    return _objectSpread({}, context, {
      assocReducer: assocReducer,
      dissocReducer: dissocReducer,
      dispatch: store.dispatch,
      getState: store.getState,
      subscribe: store.subscribe,
      runSaga: runSaga,
      supplied: true
    });
  }, [runSaga, assocReducer, dissocReducer]);
  return React__default.createElement(KLogicContext.Provider, {
    value: contextValue
  }, children);
}

var Scope = function Scope(_ref) {
  var scope = _ref.scope,
      children = _ref.children;
  var context = React.useContext(KLogicContext);
  var scopeArray = React.useMemo(function () {
    return scope.split('.');
  }, [scope]);
  var newScope = React.useMemo(function () {
    return _toConsumableArray(context.scope).concat(_toConsumableArray(scopeArray));
  }, [context.scope, scopeArray]);
  var dispatch = React.useMemo(function () {
    return kReducer.forwardTo.apply(void 0, [context.dispatch].concat(_toConsumableArray(scopeArray)));
  }, [scopeArray, context.dispatch]);
  var newContext = React.useMemo(function () {
    return _objectSpread({}, context, {
      scope: newScope,
      dispatch: dispatch
    });
  }, [newScope, dispatch]);
  return React__default.createElement(KLogicContext.Provider, {
    value: newContext
  }, children);
};

var withScope = function withScope(BaseComponent) {
  var factory = React.createFactory(BaseComponent);

  var WithScope =
  /*#__PURE__*/
  function (_Component) {
    _inheritsLoose(WithScope, _Component);

    function WithScope() {
      return _Component.apply(this, arguments) || this;
    }

    var _proto = WithScope.prototype;

    _proto.render = function render() {
      return React__default.createElement(Scope, {
        scope: this.props.scope
      }, factory(this.props));
    };

    return WithScope;
  }(React.Component);

  return WithScope;
};

var hasOwnProperty = Object.prototype.hasOwnProperty;
/**
 * inlined Object.is polyfill to avoid requiring consumers ship their own
 * https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/is
 */

function is(x, y) {
  // SameValue algorithm
  if (x === y) {
    // Steps 1-5, 7-10
    // Steps 6.b-6.e: +0 != -0
    // Added the nonzero y check to make Flow happy, but it is redundant
    return x !== 0 || y !== 0 || 1 / x === 1 / y;
  } else {
    // Step 6.a: NaN == NaN
    return x !== x && y !== y;
  }
}
/**
 * Performs equality by iterating through keys on an object and returning false
 * when any key has values which are not strictly equal between the arguments.
 * Returns true when the values of all keys are strictly equal.
 */


function shallowEqual(objA, objB, debug) {
  if (is(objA, objB)) {
    return true;
  }

  if (typeof objA !== 'object' || objA === null || typeof objB !== 'object' || objB === null) {
    return false;
  }

  var keysA = Object.keys(objA);
  var keysB = Object.keys(objB);

  if (keysA.length !== keysB.length) {
    return false;
  } // Test for A's keys different from B.


  for (var i = 0; i < keysA.length; i++) {
    if (!hasOwnProperty.call(objB, keysA[i]) || !is(objA[keysA[i]], objB[keysA[i]])) {
      if (debug) {
        console.log('shallowEqual', keysA[i], objA[keysA[i]], objB[keysA[i]]);
      }

      return false;
    }
  }

  return true;
}

var withDebug = function withDebug(BaseComponent) {
  var factory = React.createFactory(BaseComponent);

  var WithDebug =
  /*#__PURE__*/
  function (_Component) {
    _inheritsLoose(WithDebug, _Component);

    function WithDebug() {
      return _Component.apply(this, arguments) || this;
    }

    var _proto = WithDebug.prototype;

    _proto.componentDidUpdate = function componentDidUpdate(prevProps) {
      shallowEqual(prevProps, this.props, true);
    };

    _proto.render = function render() {
      return factory(this.props);
    };

    return WithDebug;
  }(React.Component);

  return WithDebug;
};

var emptyObject = {};

var useKReducer = function useKReducer(reducer, actions) {
  var context = React.useContext(KLogicContext);

  var _useState = React.useState(ramda.pathOr({}, context.scope, context.getState())),
      state = _useState[0],
      setState = _useState[1];

  var stateRef = React.useRef(state);
  var isFirstRender = React.useRef(true);
  React.useLayoutEffect(function () {
    var reducerPath = _toConsumableArray(context.scope).concat(['.']);

    context.assocReducer(reducerPath, reducer);

    var tryUpdateState = function tryUpdateState() {
      var newState = ramda.pathOr(emptyObject, context.scope, context.getState());

      if (!shallowEqual(newState, stateRef.current)) {
        setState(newState);
        stateRef.current = newState;
      }
    };

    var unsubscribe = context.subscribe(tryUpdateState);
    return function () {
      context.dissocReducer(reducerPath);
      unsubscribe();
    };
  }, []);
  var actionCreators = React.useMemo(function () {
    return bindActionCreators(actions, context.dispatch);
  }, [actions, context.dispatch]);
  var finalState = ramda.mergeDeepRight(reducer(undefined, {
    type: '@@INIT'
  }), state);

  if (isFirstRender.current) {
    isFirstRender.current = false;
  }

  return _objectSpread({}, actionCreators, finalState);
};

var asyncActionRegexp = new RegExp("^Async/(.+)/(.+)$");

var getStageLens = function getStageLens(modelDef, resource, stage, dataProp) {
  var defaultLens = ramda.compose(ramda.lensProp(dataProp), ramda.lensPath([resource, stage]));
  return modelDef[resource] && modelDef[resource][stage] ? modelDef[resource][stage] : defaultLens;
};

var buildModelLenses = function buildModelLenses(modelDef, options) {
  var dataProp = options.dataProp || 'data';
  return ramda.mapObjIndexed(function (def, resource) {
    return {
      pending: getStageLens(modelDef, resource, 'pending', dataProp),
      result: getStageLens(modelDef, resource, 'result', dataProp),
      error: getStageLens(modelDef, resource, 'error', dataProp)
    };
  }, modelDef);
};

var initModelField = function initModelField(fieldLens, defaultValue, target) {
  return ramda.compose(ramda.set(fieldLens.result, defaultValue), ramda.set(fieldLens.pending, false), ramda.set(fieldLens.error, null))(target);
};

var initModel = function initModel(modelDef, modelLenses, target) {
  return ramda.reduce(function (a, c) {
    return ramda.mergeRight(a, initModelField(modelLenses[c], modelDef[c].defaultValue || null, a));
  }, target, ramda.keys(modelDef));
};

var handleAsyncs = function handleAsyncs(modelDef, options) {
  if (options === void 0) {
    options = {};
  }

  var modelLenses = buildModelLenses(modelDef, options);
  return function (model, _ref) {
    var type = _ref.type,
        payload = _ref.payload;

    if (type === '@@INIT') {
      return initModel(modelDef, modelLenses, model);
    }

    var match = type.match(asyncActionRegexp);

    if (match) {
      var resource = match[1];
      var stage = match[2];

      if (!modelDef[resource]) {
        console.error("you dispatched async action without registering async handler. Please register a handler using handleAsyncs function.");
        return;
      }

      var resultTransform = modelDef[resource].resultTransform || ramda.identity;
      var errorTransform = modelDef[resource].errorTransform || ramda.identity;

      if (stage === 'Request') {
        return ramda.set(modelLenses[resource].pending, true, model);
      } else if (stage === 'Succeeded') {
        var m1 = ramda.set(modelLenses[resource].pending, false, model);
        return ramda.set(modelLenses[resource].result, ramda.uncurryN(2, resultTransform)(payload, model), m1);
      } else if (stage === 'Failed') {
        var _m = ramda.set(modelLenses[resource].pending, false, model);

        return ramda.set(modelLenses[resource].error, ramda.uncurryN(2, errorTransform)(payload, model), _m);
      }
    }

    return model;
  };
};

var useSagaRunner = function useSagaRunner() {
  var context = React.useContext(KLogicContext);
  var runSaga = React.useCallback(function (saga) {
    for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }

    context.runSaga.apply(context, [context.scope.join('.'), saga].concat(args));
  }, [context.scope]);
  return runSaga;
};

var asyncActionTypeName = ramda.curry(function (stage, baseType) {
  return "Async/" + baseType + "/" + stage;
});
var succeedAsyncActionName = asyncActionTypeName('Succeeded');
var failedAsyncActionName = asyncActionTypeName('Failed');
var requestedAsyncActionName = asyncActionTypeName('Request');

var createAsyncAction = function createAsyncAction(stage) {
  return function (baseType, payload) {
    return {
      type: asyncActionTypeName(stage, baseType),
      payload: payload
    };
  };
};

var requestAction = createAsyncAction('Request');
var succeededAction = createAsyncAction('Succeeded');
var failedAction = createAsyncAction('Failed');

function* asyncAction(_ref) {
  var baseType = _ref.baseType,
      fn = _ref.fn,
      args = _ref.args;

  try {
    yield effects.put(requestAction(baseType));
    var result = yield effects.call.apply(void 0, [fn].concat(_toConsumableArray(args || [])));
    yield effects.put(succeededAction(baseType, result));
    return result;
  } catch (e) {
    yield effects.put(failedAction(baseType, e));
  }
}

var fetchOnEvery = function fetchOnEvery(_ref2) {
  var actions = _ref2.actions,
      resourceKey = _ref2.resourceKey,
      fn = _ref2.fn,
      argsSelector = _ref2.argsSelector;
  return function* () {
    yield effects.takeEvery(actions, function* () {
      yield* asyncAction({
        baseType: resourceKey,
        fn: fn,
        args: []
      });
    });
  };
};

var ensureObject = ramda.unless(ramda.is(Object), ramda.objOf('value'));

var asyncAction2 =
/*#__PURE__*/
function () {
  var _ref3 = _asyncToGenerator(function* (fn, key, dispatch) {
    try {
      dispatch(requestAction(key));
      var result = yield fn();
      dispatch(succeededAction(key, result));
      return result;
    } catch (e) {
      dispatch(failedAction(key, e));
    }
  });

  return function asyncAction2(_x, _x2, _x3) {
    return _ref3.apply(this, arguments);
  };
}();

var useAsync = function useAsync(fn, key) {
  var context = React.useContext(KLogicContext);
  return React.useCallback(function () {
    asyncAction2(fn, key, context.dispatch).then();
  }, []);
};

var useSaga = function useSaga(saga, args, dependencies) {
  if (args === void 0) {
    args = [];
  }

  if (dependencies === void 0) {
    dependencies = [];
  }

  var context = React.useContext(KLogicContext);
  return React.useEffect(function () {
    context.runSaga.apply(context, [context.scope.join('.'), saga].concat(_toConsumableArray(args)));
  }, dependencies);
};

exports.handleAsyncs = handleAsyncs;
exports.asyncAction = asyncAction;
exports.fetchOnEvery = fetchOnEvery;
exports.sagaMiddleware = sagaMiddleware;
exports.KLogicProvider = KLogicProvider;
exports.KLogicContext = KLogicContext;
exports.useKReducer = useKReducer;
exports.useAsync = useAsync;
exports.Scope = Scope;
exports.withScope = withScope;
exports.withDebug = withDebug;
exports.useSaga = useSaga;
exports.shallowEqual = shallowEqual;
exports.bindActionCreators = bindActionCreators;
exports.useSagaRunner = useSagaRunner;
